import pytest
import multiprocessing
import multiprocessing.queues
import threading
import queue
import http.client
import pickle
import os
import json
import numpy as np
import time
import tempfile
from pylsl import StreamInfo, StreamOutlet

import octopus_sensing.devices.lsl_streaming as lsl_streaming
from octopus_sensing.device_coordinator import DeviceCoordinator
from octopus_sensing.common.message_creators import start_message, stop_message, terminate_message, save_message
from octopus_sensing.realtime_data_endpoint import RealtimeDataEndpoint


class RemoteEegLslDevice(threading.Thread):
    def __init__(self):
        super().__init__()
        stream_name = "FakeDeviceStream"
        stream_type = "EEG"  # or "ECG", "EDA", etc.
        self.channel_count = 8  # Number of channels to simulate (e.g., 8 EEG channels)
        self.sampling_rate = 100  # Sampling rate in Hz
        channel_format = "float32"  # Data type for the stream
        stream_id = "12345"  # Unique identifier for the stream

        # Create an LSL stream info object
        info = StreamInfo(name=stream_name,
                            type=stream_type,
                            channel_count=self.channel_count,
                            nominal_srate=self.sampling_rate,
                            channel_format=channel_format,
                            source_id=stream_id,)

        # Create an LSL outlet
        self.outlet = StreamOutlet(info)

        print(f"Streaming {stream_type} data with {self.channel_count} channels at {self.sampling_rate} Hz...")
        print("Press Ctrl+C to stop.")
        self.__running_flag = True
    
    def stop_device(self):
        self.__running_flag = False

    def run(self):
        while self.__running_flag is True:
            # Generate a random sample (replace with desired simulation logic)
            sample = np.random.randn(self.channel_count)  # Simulated data for each channel
            self.outlet.push_sample(sample)  # Send the sample over LSL
            time.sleep(1.0 / self.sampling_rate)  # Wait to maintain the sampling rate


@pytest.fixture(scope="module")
def mocked():
    original_bases = lsl_streaming.LslStreaming.__bases__[0].__bases__[0].__bases__
    lsl_streaming.LslStreaming.__bases__[0].__bases__[0].__bases__ = (threading.Thread,)
    yield None

    # Replacing back the original things
    lsl_streaming.LslStreaming.__bases__[0].__bases__[0].__bases__ = original_bases


def test_system_health(mocked):
    remote_lsl_device = RemoteEegLslDevice()
    remote_lsl_device.start()
    time.sleep(3)

    output_dir = tempfile.mkdtemp(prefix="octopus-sensing-test")
    experiment_id = 'test-exp-1'
    stimuli_id = 'sti-1'

    lsl_device_name = "FakeLslDevice"
    device = lsl_streaming.LslStreaming(lsl_device_name, "type", "EEG", 100, output_path=output_dir)

    msg_queue = queue.Queue()
    data_queue_in = queue.Queue()
    data_queue_out = queue.Queue()
    device.set_queue(msg_queue)
    device.set_realtime_data_queues(data_queue_in, data_queue_out)

    device.start()

    time.sleep(0.2)

    msg_queue.put(start_message(experiment_id, stimuli_id))
    # Allowing data collection for one second
    time.sleep(1)

    msg_queue.put(stop_message(experiment_id, stimuli_id))

    time.sleep(0.2)

    msg_queue.put(save_message(experiment_id))

    # Sending terminate and waiting for the device process to exit.
    msg_queue.put(terminate_message())
    device.join()

    remote_lsl_device.stop_device()
    remote_lsl_device.join()

    # It should save the file after receiving a TERMINATE.
    lsl_output = os.path.join(output_dir, lsl_device_name)
    filename = f"{lsl_device_name}-{experiment_id}.csv"

    assert os.path.exists(lsl_output)
    assert len(os.listdir(lsl_output)) == 1
    assert os.listdir(lsl_output)[0] == filename

    filecontent = open(os.path.join(lsl_output, filename), 'r').read()
    print(f"filecontent: {filecontent}, length={len(filecontent)}")
    assert len(filecontent) >= 375
    # TODO: Check if the triggers are there.
    # TODO: We can check data in realtime data queues as well.

def test_save_message():
    remote_lsl_device = RemoteEegLslDevice()
    remote_lsl_device.start()
    time.sleep(3)

    output_dir = tempfile.mkdtemp(prefix="octopus-sensing-test")
    experiment_id = 'test-exp-1'
    stimuli_id = 'sti-1'

    lsl_device_name = "FakeLslDevice"
    device = lsl_streaming.LslStreaming(lsl_device_name, "type", "EEG", 100, output_path=output_dir)

    msg_queue = queue.Queue()
    data_queue_in = queue.Queue()
    data_queue_out = queue.Queue()
    device.set_queue(msg_queue)
    device.set_realtime_data_queues(data_queue_in, data_queue_out)

    device.start()

    time.sleep(0.5)

    msg_queue.put(save_message(experiment_id))
    time.sleep(0.5)
    lsl_output = os.path.join(output_dir, lsl_device_name)
    print("test, lsl_output", lsl_output)
    filename = f"{lsl_device_name}-{experiment_id}.csv"
    print("test, filename", filename)

    # It should save the file after receiving a SAVE.
    assert os.path.exists(lsl_output)
    assert len(os.listdir(lsl_output)) == 1
    assert os.listdir(lsl_output)[0] == filename

    filecontent = open(os.path.join(lsl_output, filename), 'r').read()
    print(f"filecontent: {filecontent}, length={len(filecontent)}")

    # ToDo: Check the content of the file
    first_len = len(filecontent)
    print("first_len", first_len)
    assert first_len >= 375

    # Sending start and stop messages to collect data.
    msg_queue.put(start_message(experiment_id, stimuli_id))
    # Allowing data collection for one second
    time.sleep(1)

    msg_queue.put(stop_message(experiment_id, stimuli_id))

    msg_queue.put(save_message(experiment_id))
    time.sleep(1)
    filecontent = open(os.path.join(lsl_output, filename), 'r').read()
    print(f"filecontent: {filecontent}, length={len(filecontent)}")

    # ToDo: Check the content of the file
    print("second_len", len(filecontent))
    assert len(filecontent) > first_len
    assert len(filecontent) >= 375 * 2

    # Sending terminate and waiting for the device process to exit.
    msg_queue.put(terminate_message())
    device.join()

    remote_lsl_device.stop_device()
    remote_lsl_device.join()



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Foobar.
# If not, see <https://www.gnu.org/licenses/>.

from octopus_sensing.common.message import Message


class MessageType():
    '''
    Predefined message types including:

        1- START: start of a stimulus

        2- STOP: End of each stimulus

        3- TERMINATE: terminating the process of each device

        4- SAVE: Save the data in the file
    '''
    START = "START"
    STOP = "STOP"
    TERMINATE = "TERMINATE"
    SAVE = "SAVE"


def start_message(experiment_id: str, stimulus_id: str, payload=None):
    '''
    Creates a message to inform device of starting the stimulus

    Parameters
    -----------
    experiment_id: str, default: None
        A unique ID for each participant and task in the study
    
    stimulus_id: str, default: None
        A unique ID for each stimulus

    payload: Any, default: None
        the message data that can have differnt values
    

    Returns
    -------
    message: Message
        A start message

    Example
    --------
    In this example, we created a stop message. When DeviceCoordinator
    dispath this message, it will be sent to all devices in its list.
    Using this message we inform all devices that stimulus `stimulus_00` has started.

    >>> message = start_message("study_1_p10", "stimulus_00")
    >>> device_coordinator.dispatch(message)

    '''
    return Message(MessageType.START,
                   payload,
                   experiment_id=experiment_id,
                   stimulus_id=stimulus_id)


def stop_message(experiment_id: str, stimulus_id: str):
    '''
    Creates a message to inform device of stopping the stimulus


    Parameters
    -----------
    experiment_id: str, default: None
        A unique ID for each participant and task in the study
    
    stimulus_id: str, default: None
        A unique ID for each stimulus
    

    Returns
    -------
    message: Message
        A start message

    Example
    --------
    In this example, we created a start message. When DeviceCoordinator
    dispath this message, it will be sent to all devices in its list.
    Using this message we inform all devices that stimulus `stimulus_00` is finished.

    >>> message = stop_message("study_1_p10", "stimulus_00")
    >>> device_coordinator.dispatch(message)

    '''
    return Message(MessageType.STOP,
                   None,
                   experiment_id=experiment_id,
                   stimulus_id=stimulus_id)

def save_message(experiment_id: str):
    '''
    Creates a message to inform device of saving the data in the file
    This message is used when the data is saved in a continuous mode for partial save of data several times during the experiment.
    After receiving this message, the device will save the data in a file with the name and clear the data in memory.
    It will takes some time to save the data in the file, so after sending a message, wait for a short time to make sure IO is done.
    it is recommended to use this message several times in the long duration experiments to avoid losing data in case of unexpected termination of the program.
    The device will continue data recording by sending the next start message.
    
    Parameters
    -----------
    experiment_id: str, default: None
        A unique ID for each participant and task in the study

    Returns
    -------
    message: Message
        A save message

    Example
    --------
    In this example, we created a save message. When DeviceCoordinator
    dispath this message, it will be sent to all devices in its list.
    Using this message we inform all devices to save the data in file.

    >>> message = save_message("study_1_p10")
    >>> device_coordinator.dispatch(message)

    '''

    return Message(MessageType.SAVE,
                   None,
                   experiment_id=experiment_id)

def terminate_message():
    '''
    Creates a message to inform device of terminating the program
    

    Returns
    -------
    message: Message
        A terminate message

    Example
    --------
    In this example, we created a terminate message. When DeviceCoordinator
    dispath this message, it will be sent to all devices in its list.
    Using this message we inform all devices to terminate.

    >>> message = terminate_message()
    >>> device_coordinator.dispatch(message)

    '''

    return Message(MessageType.TERMINATE,
                   None)





import pytest
import multiprocessing
import multiprocessing.queues
import threading
import queue
import http.client
import pickle
import os
import json
import numpy as np
import time
import tempfile

import octopus_sensing.devices.tobiiglasses_streaming as tobiiglasses_streaming
from octopus_sensing.device_coordinator import DeviceCoordinator
from octopus_sensing.common.message_creators import start_message, stop_message, terminate_message
from octopus_sensing.realtime_data_endpoint import RealtimeDataEndpoint



class MockTobiiGlassesController():
    def __init__(self, ip_address):
        self.ip_address = ip_address
        self._is_connected = True
        self._is_streaming = False
    def get_battery_status(self):
        return {"battery_level": 100, "charging": False}
    
    def get_data(self):
        return {"mems":{'ac': {'ts': 3872249638, 's': 0, 'ac': [0.863, -9.257, 4.382]}, 'gy': {'ts': 3872260722, 's': 0, 'gy': [11.13, 89.306, 22.002]}},
                "left_eye": {'pc': {'ts': 3872276156, 's': 0, 'gidx': 15585, 'pc': [25.67, -19.62, -31.37], 'eye': 'left'}, 'pd': {'ts': 3872276156, 's': 0, 'gidx': 15585, 'pd': 3.04, 'eye': 'left'}, 'gd': {'ts': 3872276156, 's': 0, 'gidx': 15585, 'gd': [0.0338, 0.0821, 0.996], 'eye': 'left'}}, 
                "right_eye": {'pc': {'ts': 3872296141, 's': 0, 'gidx': 15586, 'pc': [-39.17, -19.39, -33.74], 'eye': 'right'}, 'pd': {'ts': 3872296141, 's': 0, 'gidx': 15586, 'pd': 2.84, 'eye': 'right'}, 'gd': {'ts': 3872276156, 's': 0, 'gidx': 15585, 'gd': [0.0754, 0.12, 0.9899], 'eye': 'right'}}, 
                "gp": {'ts': 3872276156, 's': 0, 'gidx': 15585, 'l': 59995, 'gp': [0.4823, 0.4335]}, 
                "gp3": {'ts': 3872276156, 's': 0, 'gidx': 15585, 'gp3': [38.88, 64.52, 792.89]}}
    def start_streaming(self):
        self._is_streaming = True
    def stop_streaming(self):
        self._is_streaming = False
    def close(self):
        self._is_connected = False


@pytest.fixture(scope="module")
def mocked():
    original_bases = tobiiglasses_streaming.TobiiGlassesStreaming.__bases__[0].__bases__[0].__bases__
    tobiiglasses_streaming.TobiiGlassesStreaming.__bases__[0].__bases__[0].__bases__ = (threading.Thread,)
    original_TobiiGlassesController = tobiiglasses_streaming.TobiiGlassesController
    tobiiglasses_streaming.TobiiGlassesController = MockTobiiGlassesController

    yield None

    tobiiglasses_streaming.TobiiGlassesStreaming.__bases__[0].__bases__[0].__bases__ = original_bases
    tobiiglasses_streaming.TobiiGlassesController = original_TobiiGlassesController

def test_system_health(mocked):

    output_dir = tempfile.mkdtemp(prefix="octopus-sensing-test")
    experiment_id = 'test-exp-2'
    stimuli_id = 'sti-2'

    device = \
        tobiiglasses_streaming.TobiiGlassesStreaming("192.168.71.50",
                                                     50,
                                                     output_path=output_dir,
                                                     name="tobii")
    msg_queue = queue.Queue()
    device.set_queue(msg_queue)
    realtime_data_queue_in = queue.Queue()
    realtime_data_queue_out = queue.Queue()
    device.set_realtime_data_queues(realtime_data_queue_in, realtime_data_queue_out)

    device.start()

    time.sleep(0.2)

    msg_queue.put(start_message(experiment_id, stimuli_id))
    # Allowing data collection for one second
    time.sleep(1)

    msg_queue.put(stop_message(experiment_id, stimuli_id))

    time.sleep(0.2)

    # Sending terminate and waiting for the device process to exit.
    msg_queue.put(terminate_message())
    device.join()

    # It should save the file after receiving a TERMINATE.
    tobii_output = os.path.join(output_dir, "tobii")
    filename = "tobii-{}.csv".format(experiment_id)

    assert os.path.exists(tobii_output)
    assert len(os.listdir(tobii_output)) == 1
    assert os.listdir(tobii_output)[0] == filename

    filecontent = open(os.path.join(tobii_output, filename), 'r').read()
    assert len(filecontent) >= 100
    # TODO: Check if the triggers are there.
    # TODO: We can check data in realtime data queues as well.



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.
import os
from datetime import datetime
import csv
from pydub import AudioSegment

def audio_split(log_path: str, audio_path: str):
    timestamps = []
    marks = [0]

    with open(log_path, mode='r') as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=',')
        timestamps = [(row[0], row[1]) for row in csv_reader if len(row) > 0]

    for i in range(len(timestamps) - 2):
        elapsed_time = 0
        start_time = datetime.strptime(timestamps[0][0], "%Y-%m-%d %H:%M:%S.%f")
        end_time = datetime.strptime(timestamps[i+1][0], "%Y-%m-%d %H:%M:%S.%f")

        diff = end_time - start_time
        elapsed_time = int((diff.seconds * 1000) + (diff.microseconds / 1000))
        marks.append(elapsed_time)
    
    j = 0
    for i in range(int(len(marks)*.5)):
        split_audio = AudioSegment.from_wav('Audio-px01.wav')
        split_audio = split_audio[marks[j]:marks[j+1]]
        split_audio.export(f"audio_stimuli_{timestamps[j][1]}.wav", format="wav")
        j+=2

def video_split(log_path: str, video_path: str):
    return None



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

from typing import Tuple, Any, Optional, Union
import os
import threading
import cv2
from octopus_sensing.devices.realtime_data_device import RealtimeDataDevice
from octopus_sensing.common.message_creators import MessageType
import time
from typing import Any, Dict

class CameraStreaming(RealtimeDataDevice):
    '''
    Stream and Record video data.
    If we have several stimuli, one vide file will be recorded for each stimuli.
    Device coordinator is responsible for triggerng the camera to start or stop recording.
    The content of recorded file is the recorded video between start and stop triggers

    Attributes
    ----------

    Parameters
    ----------
    name: str, default: None:
        The name of device

    output_path: str, default: output
                The path for recording files.
                Audio files will be recorded in folder {output_path}/{name}

    camera_no: int, default:0
        The camera number. Default is 0  which is defaul camera in system

    camera_path: str, default: None
        The physical path of camera device. It varies in different platforms.
        For Example in linux it can be something like this:
        `/dev/v4l/by-id/usb-046d_081b_97E6A7D0-video-index0`

    image_width: int, default: 1280
        The width of recorded frame/frames

    image_height: int, default: 720
        The height of recorded frame/frames.


    Notes
    -----
    - Only one of camera_no or camera_path should have value.

    - There is no guarantee that we can set the camera resolution.
      Because camera may not be able to support these resolution and it will change it
      based on its settings


    Example
    -----------
    Creating an instance of camera and adding it to the device coordinator.
    Device coordinator is responsible for triggerng the camera to start or stop recording

    >>> camera = CameraStreaming(camera_no=0,
    ...                          name="camera",
    ...                          output_path="./output")
    >>> device_coordinator.add_device(camera)

    See Also
    -----------
    :class:`octopus_sensing.device_coordinator`
    :class:`octopus_sensing.devices.device`
    :class:`octopus_sensing.devices.RealtimeDataDevice`

    '''
    def __init__(self, camera_no: Optional[Union[int, str]] = None,
                 camera_path: Optional[str] = None,
                 image_width: int = 1280,
                 image_height: int = 720,
                 **kwargs):
        assert (camera_no is not None) ^ (camera_path is not None), \
            "Only one of camera_no or camera_path should have value"
        super().__init__(**kwargs)
        self.output_path = os.path.join(self.output_path, self.name)
        os.makedirs(self.output_path, exist_ok=True)
        if camera_no is not None:
            self._camera_number = camera_no
        elif camera_path is not None:
            self._camera_number = os.path.realpath(camera_path)

        self._image_width = image_width
        self._image_height = image_height
        self._video_size: Tuple[int, int] = (self._image_width, self._image_height)
        self._video_capture: Any = None
        self._fps: int = 30
        self._capture_times: list = []
        self._frames: list = []
        self._counter = 0
        self._state = ""

    def _run(self):
        self._video_capture = cv2.VideoCapture(self._camera_number)
        try:
            self._video_capture.set(cv2.CAP_PROP_FRAME_WIDTH, self._image_width)
            self._video_capture.set(cv2.CAP_PROP_FRAME_HEIGHT, self._image_height)
        except Exception as error:
            # Ignoring all errors
            print(f"[{self.name}] Could not set the camera resolution. Continuing.")
            print(error)

        # There's no guarantee that we can set the camera resolution. So, we
        # re-read the settings again from the camera.
        self._video_size = (int(self._video_capture.get(cv2.CAP_PROP_FRAME_WIDTH)),
                            int(self._video_capture.get(cv2.CAP_PROP_FRAME_HEIGHT)))
        self._video_capture.read()

        if self._video_size[0] <= 0 or self._video_size[1] <= 0:
            print(f"[{self.name}] Couldn't read the video size from the camera. Trying to terminate.")
            self._video_capture.release()
            raise RuntimeError(f"[{self.name}] Couldn't read the video size from the camera.")

        recording_thread = None
        recording_event = None

        print(f"[{self.name}] Initialized video device: video size: {self._video_size}")

        while True:
            message = self.message_queue.get()
            if message is None:
                continue
            if message.type == MessageType.START:
                print(f"[{self.name}] start camera")
                if self._state == "START":
                    print("Video streaming has already started")
                else:
                    self._frames = []
                    self._capture_times = []
                    if recording_thread is not None:
                        raise RuntimeError(
                            ("[{0} device] Received two start messages. "
                            "A STOP message should be send before trying "
                            "to start a new video recording.".format(self.name)))

                    file_name = "{0}/{1}-{2}-{3}.avi".format(self.output_path,
                                                            self.name,
                                                            message.experiment_id,
                                                            str(message.stimulus_id).zfill(2))
                    print(f"[{self.name}] Starting the recording thread")
                    recording_event = threading.Event()
                    recording_event.set()
                    recording_thread = threading.Thread(
                        target=self._stream_loop, args=(file_name, recording_event), daemon=True)
                    recording_thread.start()
                    self._state = "START"

            elif message.type == MessageType.STOP:
                if self._state == "STOP":
                    print(f"[{self.name}] Video streaming has already stopped")
                else:
                    if recording_event is not None:
                        recording_event.clear()
                    recording_thread = None
                    recording_event = None
                    self._state = "STOP"

            elif message.type == MessageType.TERMINATE:
                if recording_event is not None:
                    recording_event.clear()
                recording_thread = None
                recording_event = None
                break

        print(f"[{self.name}] video terminated")
        self._video_capture.release()

    def _stream_loop(self, file_name: str, event: threading.Event):
        print(f"[{self.name}] Start stream camera")
        # It does have a VideWriter_fourcc method, but mypy can't tell.
        codec = cv2.VideoWriter_fourcc(*'XVID') # type: ignore[attr-defined]
        try:
            while self._video_capture.isOpened:
                if event.is_set():
                    ret, frame = self._video_capture.read()
                    if ret:
                        self._counter += 1
                        self._capture_times.append(time.time())
                        self._frames.append(frame)
                else:
                    fps = self._get_frame_rate()
                    print(f"[{self.name}] Recording frame per second", fps)
                    writer = cv2.VideoWriter(file_name,
                            codec,
                            fps,
                            self._video_size)
                    for frame in self._frames:
                        writer.write(frame)
                    writer.release()
                    print("Saving to file {0} is done".format(file_name))
                    break

        except Exception as error:
            print(f"[{self.name}] Error while recording video:")
            print(error)

    def _stream_loop_image(self, file_name: str, event: threading.Event):
        try:
            while self._video_capture.isOpened:
                if event.is_set():
                    ret, frame = self._video_capture.read()
                    if ret:
                        a = time.time()
                        os.makedirs(file_name[:-4], exist_ok=True)
                        image_file_name = file_name[:-4] + "/" + str(a) + ".jpg"
                        print(image_file_name)
                        cv2.imwrite(image_file_name, frame)
                else:
                    break

        except Exception as error:
            print(f"[{self.name}] Error while recording video.")
            print(error)

    def _get_realtime_data(self, duration: int) -> Dict[str, Any]:
        '''
        Returns n seconds (duration) of latest collected data for monitoring/visualizing or
        realtime processing purposes.

        Parameters
        ----------
        duration: int
            A time duration in seconds for getting the latest recorded data in realtime

        Returns
        -------
        data: Dict[str, Any]
            The keys are `data` and `metadata`.
            `data` is a list of records, or empty list if there's nothing.
            `metadata` is a dictionary of device metadata including `frame_rate` and `type`
        '''
        fps = self._get_frame_rate()

        data = self._frames[-1 * duration * fps:]
        metadata = {"frame_rate": fps,
                    "type": self.__class__.__name__}

        if len(data) == 0:
            realtime_data = {"data": [],
                             "metadata": metadata}
        else:
            one_frame_per_seconds = []
            for i in range(duration):
                if (i * fps) + 1 > len(data):
                    break
                one_frame_per_seconds.append(data[i * fps])

            realtime_data = {"data": one_frame_per_seconds,
                             "metadata": metadata}
        return realtime_data

    def _get_frame_rate(self):
        time_diff = 1
        i = 1
        while time_diff < 5 and len(self._capture_times) > i:
            time_diff = self._capture_times[-1] - self._capture_times[-(i+1)]
            i += 1
        fps = int(i/time_diff)

        return fps



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

import sys
import time
import traceback
import multiprocessing
import queue
import pickle
from typing import List, Any, Tuple, Dict, Optional

from octopus_sensing.devices.device import Device
from octopus_sensing.devices.realtime_data_device import RealtimeDataDevice
from octopus_sensing.common.message import Message
from octopus_sensing.common.message_creators import terminate_message

QueueType = multiprocessing.queues.Queue


class RealtimeDataCache:
    def __init__(self) -> None:
        self._time = time.time_ns()
        self._cached_data: Dict[str, List[Any]] = {}

    def get_cache(self):
        '''Returns None if cache is not available or expired'''
        # 100 ms
        if time.time_ns() - self._time > 100000000:
            return None
        return self._cached_data

    def cache(self, data: Dict[str, List[Any]]):
        self._cached_data = data
        self._time = time.time_ns()


class DeviceCoordinator:
    '''
    Device coordinator is responsible for coordination, like start recording data
    from all devices at once, stop recording, triggering (marking data at point),
    and terminating devices.
    When a device is added to the device coordinator, it will be initialized and prepared for data recording.
    '''

    def __init__(self) -> None:
        self.__devices: Dict[str, Device] = {}
        self.__queues: List[QueueType] = []
        self.__device_counter: int = 0
        self.__realtime_data_queues: List[Tuple[QueueType, QueueType, Device]] = [
        ]
        self.__realtime_data_cache = RealtimeDataCache()

    def __get_device_id(self) -> str:
        '''
        Generate an ID for devices that do not have name

        Returns
        -----------
        str
            an ID for device
        '''

        self.__device_counter += 1
        device_id = "device_{0}".format(self.__device_counter)
        return device_id

    def get_devices(self) -> List[Device]:
        '''
        Return a list of added devices

        Returns
        -----------
        List[Device]
            List of device objects
        '''

        return list(self.__devices.values())

    def add_device(self, device: Device) -> None:
        '''
        Adds new device to the coordinator and starts it

        Parameters
        ----------
        device: Device
            a device object

        Example
        --------
        >>> my_shimmer = Shimmer3Streaming(name="Shimmer3_sensor", output_path="./output")
        >>> device_coordinator.add_device(my_shimmer)

        See Also
        ---------
        add_devices: Adds a list of new devices to the coordinator and starts them
        '''
        assert isinstance(device, Device)

        if device.name is None:
            device.name = self.__get_device_id()
        if device.name in self.__devices.keys():
            raise RuntimeError("This device already has been added")

        self.__devices[device.name] = device
        msg_queue: QueueType = multiprocessing.Queue()
        device.set_queue(msg_queue)
        self.__queues.append(msg_queue)

        self.__set_realtime_data_queues(device)

        device.start()

    def add_devices(self, devices: List[Device]) -> None:
        '''
        Adds a list of new devices to the coordinator and starts them

        Parameters
        ----------
        devices: List[Device]
            a list of device objects

        Example
        --------
        >>> my_shimmer = Shimmer3Streaming(name="Shimmer3_sensor",
                                           output_path="./output")
        >>> device_coordinator.add_devices([my_shimmer])

        See Also
        ---------
        add_device: Adds new device to the coordinator and starts it
        '''
        for device in devices:
            self.add_device(device)

    def dispatch(self, message: Message) -> None:
        '''
        dispatches a new message to all devices

        Parameters
        ----------
        message: Message
            a message object

        Example
        --------
        >>> device_coordinator.dispatch(start_message(experiment_id,
                                                      stimuli_id))
        '''

        for message_queue in self.__queues:
            message_queue.put(message)

    def health_check(self):
        '''
        Checks if all the devices are okay.

        Returns
        -----------
        boolean
            True if all are healthy, False otherwise
        '''
        for device_name, device, in self.__devices.items():
            if not device.is_alive():
                print("device [{0}] is not alive anymore.".format(device_name))
                return False
        return True

    def terminate(self):
        '''
        sends terminate message to all devices and terminate all processes

        Example
        --------
        >>> device_coordinator.terminate()
        '''
        self.dispatch(terminate_message())
        for item in self.__devices.values():
            item.join()

    def get_realtime_data(self, duration: int, device_list: Optional[List[str]]) -> Dict[str, List[Any]]:
        '''
        Returns latest collected data from all devices.
        Device's data can be anything, depending on the device itself.

        Parameters
        ----------
        duration: int
            a time duration for getting realtime data in seconds

        device_list: List[str]
            a list of device names. Only devices in this list will be monitored or processed in realtime

        Returns
        ---------
        data : dict[str, list[any]]
            The keys are device names and values are collected data from the device

        Note
        ----
        This method is being used for getting data in real-time for monitoring or realtime processing

        '''
        cached = self.__realtime_data_cache.get_cache()
        if cached:
            return cached

        out_queues: List[Tuple[QueueType, str]] = []
        # Putting request for all devices, then collecting them all, for performance reasons.
        for in_q, out_q, device in self.__realtime_data_queues:
            if device_list is None or device.name in device_list:
                try:
                    # The sub-process won't use the data we put in the queue. It's just a signal.
                    in_q.put(str(duration), timeout=0.1)
                    out_queues.append((out_q, device.name))
                except queue.Full:
                    print("Could not put realtime data request for {0} device.".format(
                        device.name), file=sys.stderr)
                    traceback.print_exc()

        result: Dict[str, List[Any]] = {}
        for out_q, device_name in out_queues:
            try:
                records = pickle.loads(out_q.get(timeout=0.1))
                # We ensured device has a name in the add_device, ignoring it here.
                result[device_name] = records  # type: ignore
            except (queue.Empty, pickle.PickleError):
                print("Could not read realtime data from {0} device".format(
                    device_name), file=sys.stderr)
                traceback.print_exc()

        self.__realtime_data_cache.cache(result)
        return result

    def __set_realtime_data_queues(self, device: Device) -> None:
        if isinstance(device, RealtimeDataDevice):
            in_q: QueueType = multiprocessing.Queue()
            out_q: QueueType = multiprocessing.Queue()
            device.set_realtime_data_queues(in_q, out_q)
            self.__realtime_data_queues.append((in_q, out_q, device))



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, Dict, Any, List

from octopus_sensing.common.endpoint_base import EndpointBase


class RealtimeDataEndpoint(EndpointBase):

    def __init__(self, device_coordinator):
        super().__init__(endpoint_name="RealtimeDataEndpoint-Thread",
                         port=9330, get_callback=self._get_handler)
        self._device_coordinator = device_coordinator

    def _get_handler(self, request_reader, query_params: Dict[str, List[str]]):
        duration: int = int(query_params.get('duration', ["3"])[0])

        device_list_string = query_params.get('device_list', [""])[0]
        device_list: Optional[List[str]] = None
        if device_list_string != "":
            device_list = device_list_string.split(',')
        return self._device_coordinator.get_realtime_data(duration, device_list)



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

from datetime import datetime
import time
import os
import threading
import csv
import numpy as np
from typing import List, Optional, Dict, Any

from octopus_sensing.common.message_creators import MessageType
from octopus_sensing.devices.realtime_data_device import RealtimeDataDevice
from octopus_sensing.devices.common import SavingModeEnum

from libtobiiglassesctrl import TobiiGlassesController


class TobiiGlassesStreaming(RealtimeDataDevice):
    '''
    Manage Tobii Glasses streaming

    Attributes
    ----------
    device_ip
        Device IP address. (e.g.192.168.71.50)

    sampling_rate
        the sampling rate for recording data


    name
        device name
        This name will be used in the output path to identify each device's data

    output_path
        The path for recording files.
        Recorded files will be recorded in folder {output_path}/{name}

    saving_mode
        The way of saving data. It saves data continiously in a file
        or saves data which are related to various stimulus in separate files.
        default is SavingModeEnum.CONTINIOUS_SAVING_MODE
        SavingModeEnum is [CONTINIOUS_SAVING_MODE, SEPARATED_SAVING_MODE]
    
    Notes
    -----
    This class is used for reading data from Tobii Pro Glasses 2.
    It uses the `libtobiiglassesctrl` library to connect to the device and read data.
    The output file is in CSV format and its columns are as follows:
    [ac_ts, ac_x, ac_y, ac_z,
    gy_ts, gy_x, gy_y, gy_z,
    right_eye_pc_ts, right_eye_pc_x, right_eye_pc_y, right_eye_pc_z,
    right_eye_pd_ts, right_eye_pd,
    right_eye_gd_ts, right_eye_gd_x, right_eye_gd_y, right_eye_gd_z,
    left_eye_pc_ts, left_eye_pc_x, left_eye_pc_y, left_eye_pc_z,
    left_eye_pd_ts, left_eye_pd,
    left_eye_gd_ts, left_eye_gd_x, left_eye_gd_y, left_eye_gd_z,
    gp_ts, gp_l, gp_x, gp_y,
    gp3_ts, gp3_x, gp3_y, gp3_z,
    timestamp,
    trigger]
    ts: timestamp for each sensor
    ac: accelometer data
    gy: gyroscope data  
    pc: pupil center
    pd: pupil diameter
    gd: gaze direction
    gp: gaze position
    gp3: 3D gaze position
    timestamp: the time when the data is recorded
    The `trigger` column is used to identify the stimuli marker

    See Also
    -----------
    :class:`octopus_sensing.device_coordinator`
    :class:`octopus_sensing.devices.device`

    Examples
    ---------
    Here is an example of using Tobii pro glasses 2 for reading data

    >>> params.serial_port = "/dev/ttyUSB0"
    >>> my_tobiiglasses =
    ...       TobiGlassesStreaming("192.168.71.50",
    ...                            50,
    ...                            name="tobii_glasses",
    ...                            output_path="./output",
    ...                            saving_mode=SavingModeEnum.CONTINIOUS_SAVING_MODE)

    '''
    def __init__(self,
                 device_ip: str = "192.168.71.50",
                 sampling_rate: int = 50,
                 saving_mode: int=SavingModeEnum.CONTINIOUS_SAVING_MODE,
                 name: Optional[str] = None,
                 output_path: str = "output"):
        super().__init__(name=name, output_path=output_path)

        self._saving_mode = saving_mode
        self._stream_data: List[float] = []
        self.sampling_rate = sampling_rate

        self._board = None
        self._device_ip = device_ip
        self._terminate = False
        self._trigger = None
        self._experiment_id = None
        self.__loop_thread: Optional[threading.Thread] = None
        self._controller = None

        self.output_path = os.path.join(self.output_path, self.name)
        os.makedirs(self.output_path, exist_ok=True)
        self._state = ""

    def _run(self):
        self._controller = TobiiGlassesController(self._device_ip)
        print("TobiiGlasses streaming: Connecting to the device...")
        print(self._controller.get_battery_status())
        
        self.__loop_thread = threading.Thread(target=self._stream_loop)
        self.__loop_thread.start()

        while True:
            message = self.message_queue.get()

            if not self.__loop_thread.is_alive():
                print("TobiiGlasses streaming: The streaming thread is dead. Terminating.")
                break

            if message is None:
                continue

            if message.type == MessageType.START:
                if self._state == "START":
                    print("TobiiGlasses streaming has already recorded the START triger")
                else:
                    print("TobiiGlasses start")
                    self.__set_trigger(message)
                    self._experiment_id = message.experiment_id
                    self._state = "START"
            elif message.type == MessageType.STOP:
                if self._state == "STOP":
                    print("Tobii glasses streaming has already recorded the STOP triger")
                else:
                    print("Tobii glasses stop")
                    if self._saving_mode == SavingModeEnum.SEPARATED_SAVING_MODE:
                        self._experiment_id = message.experiment_id
                        file_name = \
                            "{0}/{1}-{2}-{3}.csv".format(self.output_path,
                                                        self.name,
                                                        self._experiment_id,
                                                        message.stimulus_id)
                        self._save_to_file(file_name)
                        self._stream_data = []
                    else:
                        self._experiment_id = message.experiment_id
                        self.__set_trigger(message)
                    self._state = "STOP"
            elif message.type == MessageType.SAVE:
                if self._saving_mode == SavingModeEnum.CONTINIOUS_SAVING_MODE:
                    self._experiment_id = message.experiment_id
                    file_name = \
                        "{0}/{1}-{2}.csv".format(self.output_path,
                                                 self.name,
                                                 self._experiment_id)
                    self._save_to_file(file_name)
                    self._stream_data = []
            elif message.type == MessageType.TERMINATE:
                self._terminate = True
                if self._saving_mode == SavingModeEnum.CONTINIOUS_SAVING_MODE:
                    file_name = \
                        "{0}/{1}-{2}.csv".format(self.output_path,
                                                 self.name,
                                                 self._experiment_id)
                    self._save_to_file(file_name)
                break

        self._controller.stop_streaming()
        self._controller.close()
        self.__loop_thread.join()

    def _stream_loop(self):
        self._controller.start_streaming()
        while True:
            if self._terminate is True:
                break
            data = self.__safe_get(self._controller.get_data())
            
            if self._trigger is not None:
                data.append(self._trigger)
                self._trigger = None

            self._stream_data.append(np.array(data))
            time.sleep(1/self.sampling_rate)

    def __set_trigger(self, message):
        '''
        Takes a message and set the trigger using its data

        Parameters
        ----------
        message: Message
            a message object
        '''
        self._trigger = \
            "{0}-{1}-{2}".format(message.type,
                                 message.experiment_id,
                                 str(message.stimulus_id).zfill(2))

    def _save_to_file(self, file_name):
        print("Saving {0} to file {1}".format(self._name, file_name))
        header = ["ac_ts", "ac_x", "ac_y", "ac_z",
                  "gy_ts", "gy_x", "gy_y", "gy_z",
                  "left_eye_pc_ts", "left_eye_pc_x", "left_eye_pc_y", "left_eye_pc_z",
                  "left_eye_pd_ts", "left_eye_pd",
                  "left_eye_gd_ts", "left_eye_gd_x", "left_eye_gd_y", "left_eye_gd_z",
                  "right_eye_pc_ts", "right_eye_pc_x", "right_eye_pc_y", "right_eye_pc_z",
                  "right_eye_pd_ts", "right_eye_pd",
                  "right_eye_gd_ts", "right_eye_gd_x", "right_eye_gd_y", "right_eye_gd_z",
                  "gp_ts", "gp_l", "gp_x", "gp_y",
                  "gp3_ts", "gp3_x", "gp3_y", "gp3_z",
                  "timestamp",
                  "trigger"]
        with open(file_name, 'a') as csv_file:
            writer = csv.writer(csv_file)
            if os.stat(file_name).st_size == 0:
                print("TobiiGlassesStreaming: file created")
                writer.writerow(header)
                csv_file.flush()
            print("TobiiGlassesStreaming: file already exists, appending data")
            for row in self._stream_data:
                writer.writerow(row)
                csv_file.flush()
        print("Saving {0} to file {1} is done".format(self._name, file_name))


    def _get_realtime_data(self, duration: int) -> Dict[str, Any]:
        '''
        Returns n seconds (duration) of latest collected data for monitoring/visualizing or
        realtime processing purposes.

        Parameters
        ----------
        duration: int
            A time duration in seconds for getting the latest recorded data in realtime

        Returns
        -------
        data: Dict[str, Any]
            The keys are `data` and `metadata`.
            `data` is a list of records, or empty list if there's nothing.
            `metadata` is a dictionary of device metadata including `sampling_rate` and `channels` and `type`

        '''
        # Last seconds of data

        data = self._stream_data[-1 * duration * self.sampling_rate:]
        metadata = {"sampling_rate": self.sampling_rate,
                    "type": self.__class__.__name__}

        realtime_data = {"data": data,
                  "metadata": metadata}
        return realtime_data


    def __safe_get(self, data):
        '''
        Safely get nested dictionary values with fallback to None.
        Parameters
        ----------
        data: dict
            The data dictionary to extract values from.
        Returns
        -------
        flat_data: list
            A flat list of values extracted from the nested dictionary.
        '''
        flat_data = []
        mems = data.get("mems", [None]*8)
        if isinstance(mems, dict):
            mems_ac = mems.get("ac", [None]*4)
            if isinstance(mems_ac, dict):
                flat_data.extend([mems_ac.get("ts", None), *mems_ac.get("ac", [None]*3)])
            mems_gy = mems.get("gy", [None]*4)
            if isinstance(mems_gy, dict):
                flat_data.extend([mems_gy.get("ts", None), *mems_gy.get("gy", [None]*3)])
        left_eye = data.get("left_eye", [None]*10)
        if isinstance(left_eye, dict):
            left_pc = left_eye.get("pc", [None]*4)
            if isinstance(left_pc, dict):
                flat_data.extend([left_pc.get("ts", None), *left_pc.get("pc", [None]*3)])

            left_pd = left_eye.get("pd", [None, None])
            flat_data.extend([left_pd.get("ts", None), left_pd.get("pd", None)])

            left_gd = left_eye.get("gd", [None]*4)
            if isinstance(left_gd, dict):
                flat_data.extend([left_gd.get("ts", None), *left_gd.get("gd", [None]*3)]) 
        right_eye = data.get("right_eye", [None]*10)
        if isinstance(right_eye, dict):
            right_pc = right_eye.get("pc", [None]*4)
            if isinstance(right_pc, dict):
                flat_data.extend([right_pc.get("ts", None), *right_pc.get("pc", [None]*3)])

            right_pd = right_eye.get("pd", [None, None])
            flat_data.extend([right_pd.get("ts", None), right_pd.get("pd", None)])

            right_gd = right_eye.get("gd", [None]*4)
            if isinstance(right_gd, dict):
                flat_data.extend([right_gd.get("ts", None), *right_gd.get("gd", [None]*3)])
        gp = data.get("gp", [None]*4)
        if isinstance(gp, dict):
            flat_data.extend([gp.get("ts", None), gp.get("l", None), *gp.get("gp", [None]*2)])
        gp3 = data.get("gp3", [None]*4)
        if isinstance(gp3, dict):
            flat_data.extend([gp3.get("ts", None), *gp3.get("gp3", [None]*3)])
        flat_data.append(time.time())

        return flat_data
    



import queue
import os
import time
import tempfile
import threading

from octopus_sensing.devices.testdevice_streaming import TestDeviceStreaming
from octopus_sensing.common.message_creators import start_message, stop_message, terminate_message


def test_system_health():
    print("test device")
    output_dir = tempfile.mkdtemp(prefix="octopus-sensing-test")
    experiment_id = 'test-exp-2'
    stimuli_id = 'sti-2'

    device = \
        TestDeviceStreaming(12,
                            name="test_device",
                            output_path=output_dir)
    msg_queue = queue.Queue()
    device.set_queue(msg_queue)
    realtime_data_queue_in = queue.Queue()
    realtime_data_queue_out = queue.Queue()
    device.set_realtime_data_queues(realtime_data_queue_in, realtime_data_queue_out)

    device_thread = threading.Thread(target=device._run)
    device_thread.start()
    print("after start")

    time.sleep(0.2)

    msg_queue.put(start_message(experiment_id, stimuli_id))
    # Allowing data collection for one second
    time.sleep(1)

    msg_queue.put(stop_message(experiment_id, stimuli_id))

    time.sleep(0.2)

    # Sending terminate and waiting for the device process to exit.
    msg_queue.put(terminate_message())
    device_thread.join()

    # It should save the file after receiving a TERMINATE.
    test_device_output = os.path.join(output_dir, "test_device")
    filename = "test_device-{}.csv".format(experiment_id)

    assert os.path.exists(test_device_output)
    assert len(os.listdir(test_device_output)) == 1
    assert os.listdir(test_device_output)[0] == filename

    filecontent = open(os.path.join(test_device_output, filename), 'r').read()
    assert len(filecontent) >= 375
    # TODO: Check if the triggers are there.
    # TODO: We can check data in realtime data queues as well.



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

import threading
import csv
import sys
import os
from typing import List, Optional, Any, Dict
from pylsl import StreamInlet, resolve_byprop

from octopus_sensing.common.message_creators import MessageType
from octopus_sensing.devices.realtime_data_device import RealtimeDataDevice
from octopus_sensing.devices.common import SavingModeEnum


class LslStreaming(RealtimeDataDevice):
    '''
    Get and Record data from a LSL stream.

    Attributes
    ----------

    Parameters
    ----------
    
    name: str
          device name
          This name will be used in the output path to identify each device's data

    stream_property_type: str
            It uses the property info to resolve a device from network. This the info provided by the LSL streamer
            you want to read. For example, stream_property_type='name' and stream_property_value='EEG' will try to
            find any device that it's name is 'EEG' and is streaming in the current network.

    stream_property_value: str
            See stream_property_type
    
    output_path: str, optional
            The path for recording files.
            Recorded file/files will be in folder {output_path}/{name}

    Example
    -------
    Creating an instance of LSL streaming recorder and adding it to the device coordinator.
    Device coordinator is responsible for triggerng the device recorder to start or stop recording

    >>> lsl_device = LslStreaming("mbtrain",
    ...                             "name",
    ...                             "EEG",
                                    250,
    ...                             output_path="output")
    >>> device_coordinator.add_device(lsl_device)

    See Also
    -----------
    :class:`octopus_sensing.device_coordinator`
    :class:`octopus_sensing.devices.device`

    '''

    def __init__(self,
                 name: str,
                 stream_property_type: str,
                 stream_property_value: str,
                 sampling_rate: int,
                 output_path: str = "output",
                 saving_mode: int=SavingModeEnum.CONTINIOUS_SAVING_MODE,
                 channels: Optional[list] = None,
                 **kwargs):
        super().__init__(**kwargs)
        self._name = name
        self._stream_property_type = stream_property_type
        self._stream_property_value = stream_property_value
        self._stream_data: List[float] = []
        self._loop_thread: Optional[threading.Thread] = None
        self._terminate = False
        self._state = ""
        self._experiment_id = None
        self._stream = None
        self._inlet = None
        self.sampling_rate = sampling_rate
        self.channels = channels
        self._trigger = None
        self._saving_mode = saving_mode

        self.output_path = os.path.join(output_path, self._name)
        os.makedirs(self.output_path, exist_ok=True)

    def _run(self):
        '''
        Listening to the message queue and manage messages
        '''
        self._loop_thread = threading.Thread(target=self._stream_loop)
        self._loop_thread.start()

        while True:
            message = self.message_queue.get()
            if message is None:
                continue

            if message.type == MessageType.START:
                if self._state == "START":
                    print(f"LSL Device: '{self.name}' has already started.")
                else:
                    print(f"LSL Device: '{self.name}' started.")
                    self._experiment_id = message.experiment_id
                    self.__set_trigger(message)
                    self._state = "START"

            elif message.type == MessageType.STOP:
                if self._state == "STOP":
                    print(f"LSL Device '{self.name}' has already stopped.")
                else:
                    print(f"LSL Device '{self.name}' stopped.")
                    if self._saving_mode == SavingModeEnum.SEPARATED_SAVING_MODE:
                        self._experiment_id = message.experiment_id
                        file_name = \
                            "{0}/{1}-{2}-{3}.csv".format(self.output_path,
                                                        self.name,
                                                        self._experiment_id,
                                                        message.stimulus_id)
                        self._save_to_file(file_name)
                        self._stream_data = []
                    else:
                        self._experiment_id = message.experiment_id
                        self.__set_trigger(message)
                    self._state = "STOP"              
            elif message.type == MessageType.SAVE:
                if self._saving_mode == SavingModeEnum.CONTINIOUS_SAVING_MODE:
                    self._experiment_id = message.experiment_id
                    file_name = \
                        "{0}/{1}-{2}.csv".format(self.output_path,
                                                 self.name,
                                                 self._experiment_id)
                    self._save_to_file(file_name)
                    self._stream_data = []
            elif message.type == MessageType.TERMINATE:
                self._terminate = True
                if self._saving_mode == SavingModeEnum.CONTINIOUS_SAVING_MODE:
                    file_name = \
                        "{0}/{1}-{2}.csv".format(self.output_path,
                                                 self.name,
                                                 self._experiment_id)
                    self._save_to_file(file_name)
                break

        self._loop_thread.join()

    def _stream_loop(self):
        # self._stream = resolve_stream(self._device_type, self.device)
        self._stream = resolve_byprop(self._stream_property_type, self._stream_property_value, timeout=5)
        if self._stream is None or len(self._stream) == 0:
            raise RuntimeError(f"Couldn't resolve an LSL stream with {self._stream_property_type}={self._stream_property_value}")
        self._inlet = StreamInlet(self._stream[0])

        while True:
            if self._terminate is True:
                break
            sample, timestamp = self._inlet.pull_sample(timeout=0.2)
            if sample is not None:
                sample.append(timestamp)
                if self._trigger is not None:
                    sample.append(self._trigger)
                    self._trigger = None
                self._stream_data.append(sample)

    def __set_trigger(self, message):
        '''
        Takes a message and set the trigger using its data

        Parameters
        ----------
        message: Message
            a message object
        '''
        # Add the trigger to the data
        self._trigger = \
            "{0}-{1}-{2}".format(message.type,
                                 message.experiment_id,
                                 str(message.stimulus_id).zfill(2))

    def _save_to_file(self, file_name):
        print("Saving {0} to file {1}".format(self._name, file_name))
        if not os.path.exists(file_name):
            csv_file = open(file_name, 'a')
            writer = csv.writer(csv_file)
            csv_file.flush()
            csv_file.close()
        
        with open(file_name, 'a') as csv_file:
            writer = csv.writer(csv_file)
            for row in self._stream_data:
                writer.writerow(row)
                csv_file.flush()
        print("Saving {0} to file {1} is done".format(self._name, file_name))

    def _get_realtime_data(self, duration: int) -> Dict[str, Any]:
            '''
            Returns n seconds (duration) of latest collected data for monitoring/visualizing or
            realtime processing purposes.

            Parameters
            ----------
            duration: int
                A time duration in seconds for getting the latest recorded data in realtime

            Returns
            -------
            data: Dict[str, Any]
                The keys are `data` and `metadata`.
                `data` is a list of records, or empty list if there's nothing.
                `metadata` is a dictionary of device metadata including `sampling_rate` and `channels` and `type`

            '''
            # Last seconds of data

            data = self._stream_data[-1 * duration * self.sampling_rate:]
            metadata = {"sampling_rate": self.sampling_rate,
                        "channels": self.channels,
                        "type": self.__class__.__name__}

            realtime_data = {"data": data,
                    "metadata": metadata}
            return realtime_data
